//******************************************************************************
// jacobi-seq.c
//
// Jacobi iteration for the Laplacian equation Laplace u = f,
// sequential version with Finite-Differences (FD). Consistency order is
// 2, Jacobi converges but extremly slow for n >= 64 (eventually use euklid.
// instead max. norm)
//
// 06.12.2011
//******************************************************************************
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>

const double Pi = 3.1415926535897932385;

// corner up right
const double eor = 1.0;

// soruce f
double f(double x1, double x2)
{
  return 2.0 * Pi * Pi * sin(Pi*x1) * sin(Pi*x2);
}

// evaluation of the analytical solution
double aSol(double x1, double x2, double sol, double err)
{
  return fabs(sol - sin(Pi*x1) * sin(Pi*x2));
}

// evaluation of the error in the maximum norm on all grid points
double maxNorm(double* sol, long n)
{
  long i, j;
  double maxErr = 0.0;
  const double h  = eor / (n - 1);

  for (i=0; i<n; i++)
  {
    for (j=0; j<n; j++)
    {
      const double x1 = i * h;
      const double x2 = j * h;
      double tmp = fabs(sol[i*n+j] - sin(Pi*x1) * sin(Pi*x2));
      if (maxErr < tmp)
        maxErr = tmp;
    }
  }

  return maxErr;
}

// maximum norm of the residual
double maxNormRes(const double* const x, const double* const b, double** A, long m)
{
  double max = 0.0;
  long i,j;

  for (i=0; i<m; i++)
  {
    double res = b[i];
    for (j=0; j<m; j++)
      res -= A[i][j] * x[j];
    if (max < fabs(res))
      max  = fabs(res);
  }

  return max;
}

// initialisation of matrix and vectors
void init(double* x, double* b, double** A, long m, long n)
{
  const double h  = eor / (n - 1);
  const double h2 = h * h;
  long i, j, k;

  // iterate over all grid points
  for (i=0; i<n; i++)
  {
    for (j=0; j<n; j++)
    {
      // row index
      const long zi = i * n + j;

      // initialize the solution vector x with 0
      x[zi] = 0.0;

      // initialize the Last vector b with f(x,y)
      const double x1 = i * h;
      const double x2 = j * h;
      b[zi] = f(x1, x2);

      // generate the matrix row for grid point (i,j)
      for (k=0; k<m; k++)
      {
        // reset row 
        A[zi][k] = 0.0;

        // diagonal entry
        A[zi][zi] = 4.0 / h2;

        // Off-diagonal entry
        if ((k == zi-1) || (k == zi+1) || (k == zi-n) || (k == zi+n))
          A[zi][k] = -1.0 / h2;

        // set boundary conditions
        if ((i == 0) || (i == n-1) || (j == 0) || (j == n-1))
        {
          x[zi] = 0.0;
          b[zi] = 0.0;

          A[zi][k]  = 0.0;
          A[zi][zi] = 1.0;
        }
      }
    }
  }
}

// output a matrix
void outputMatrix(double** A, long m)
{
  long i, j;
  for (i=0; i<m; i++)
  {
    for (j=0; j<m; j++)
    {
      fprintf(stdout, "%.4e  ", A[i][j]);
    }
    fprintf(stdout, "\n");
  }
  fprintf(stdout, "\n\n");
}

// output a vector
void output(double* v, long n)
{
  long i, j;
  const double h  = eor / (n - 1);
  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
    {
      const double x1 = i * h;
      const double x2 = j * h;
      fprintf(stdout, "%.4e  %.4e  %.4e\n", x1, x2, v[i*n+j]);
    }
  fprintf(stdout, "\n\n");
}

// Jacobi step
void jacobi(double* xneu, const double* const xalt, const double* const b, double** A, long m)
{
  long i=0, j=0;

  for (i=0; i<m; ++i)
  {
    xneu[i] = 0.0;

    // calculate the scalar product
    for (j=0; j<m; ++j)
    {
      xneu[i] += A[i][j] * xalt[j];
    }

    // subtract the case i = j
    xneu[i] -= A[i][i] * xalt[i];

    // finish of the Jacobi step
    xneu[i] *= -1.0;
    xneu[i] += b[i];
    if (A[i][i] == 0)
    {
      fprintf(stderr, "Division by 0\n.");
      exit (1);
    }
    else
      xneu[i] /= A[i][i];
  }
}

//*****************************************************************************
// main
//*****************************************************************************
int main(int argc, char** argv)
{
  const long maxIt = 10000; // max. iterations of the method
  long it;                  // loop counter
  long n;                   // problem size in one direction

  double*  x;               // unknowns
  double*  y;               // old unknowns
  double*  b;               // right side
  double** A;               // matrix

  const double eps = 1e-5;  // termination condition
  double diff;              // maximum norm of the distance of two solutions
  double t = 0.0;           // for the time measurement

  // read the length of the unknown vector x
  if (argc != 2)
  {
    fprintf(stderr, "Syntax: %s <n>\n", argv[0]);
    exit (1);
  }
  n = atoi(argv[1]);

  // allocate memory for A, x, y and b
  const long m = n * n;
  x = (double *)  malloc(sizeof(double)  * m); // old vector x
  y = (double *)  malloc(sizeof(double)  * m); // new vector y
  b = (double *)  malloc(sizeof(double)  * m); // right hand side b
  A = (double **) malloc(sizeof(double*) * m); // rows of matrix A
  long i;
  for (i=0; i<m; i++)
    A[i] = (double*) malloc(sizeof(double) * m);

  if (A == NULL || b == NULL || y == NULL || x == NULL)
  {
    fprintf(stderr, "Could not allocate memory\n");
    return 1;
  }

  // initialise A, b and x
  init(x, b, A, m, n);

  // Jacobi steps
  t = MPI_Wtime();
  for (it=1; it<((long) (maxIt/2)+1); it++)
  {
    // trick of copying x^{m} = y (new vector into old vector)
    // to save vector: 2 Jacobi steps, solution x^(m) is then alternating
    // in x or y. this means we make in total 2*it Jacobi steps
    jacobi(y, x, b, A, m);
    jacobi(x, y, b, A, m);

    // concergence check and end ot time measurement
    diff = maxNormRes(x, b, A, m);
    fprintf(stdout, "Norm(res): %.16e\n", diff);
    if (diff < eps)
    {
      t = MPI_Wtime() - t;
      //fprintf(stdout, "Iter: %ld, t: %f s, Norm(res): %.16e\n", (2*it), t, diff);
      break;
    }
  }

  // output of final solution vector and error in max norm
  output(x, n);
  printf("Max. Error: %e\n", maxNorm(x, n));

  // free memory 
  free(x);
  free(y);
  free(b);
  for (i=0; i<m; i++)
    free(A[i]);
  free(A);

  return (0);
}
